/*
 * dhcpcd-qt
 * Copyright 2014-2017 Roy Marples <roy@marples.name>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef DHCPCD_WI_H
#define DHCPCD_WI_H

#include <QAction>
#include <QObject>

#include "dhcpcd.h"

class QMenu;
class QSocketNotifier;
class QTimer;
class QWidgetAction;

class DhcpcdQt;
class DhcpcdSsid;
class DhcpcdSsidMenu;

class DhcpcdWi : public QObject
{
	Q_OBJECT

public:
	DhcpcdWi(DhcpcdQt *dhcpcdQt, DHCPCD_WPA *wpa);
	~DhcpcdWi();
	DHCPCD_WPA *getWpa();
	bool open();
	void close();

	DHCPCD_WI_SCAN *getScans();
	bool setScans(DHCPCD_WI_SCAN *scans);

	void createMenu(QMenu *parent);
	QMenu *createIfMenu(QMenu *parent);

private slots:
	void dispatch();
	void ping();
	void connectSsid(DHCPCD_WI_SCAN *scan);
	void scan();
	void menuHidden();
	void menuShown();

private:
	DhcpcdQt *dhcpcdQt;
	DHCPCD_WPA *wpa;
	DHCPCD_WI_SCAN *scans;
	DhcpcdSsid *ssid;

	QSocketNotifier *notifier;
	QTimer *pingTimer;
	QTimer *scanTimer;

	QMenu *menu;
	void createMenuItem(QMenu *menu, DHCPCD_WI_SCAN *scan,
	    QAction *before = NULL);
	void createMenu1(QMenu *parent);
};

#endif
